# MLflow webhook headers
WEBHOOK_SIGNATURE_HEADER = "X-MLflow-Signature"
WEBHOOK_TIMESTAMP_HEADER = "X-MLflow-Timestamp"
WEBHOOK_DELIVERY_ID_HEADER = "X-MLflow-Delivery-Id"

# Webhook signature version
WEBHOOK_SIGNATURE_VERSION = "v1"
