import databricks_pb2 as _databricks_pb2
from scalapb import scalapb_pb2 as _scalapb_pb2
import unity_catalog_oss_messages_pb2 as _unity_catalog_oss_messages_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import service as _service
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor

class UnityCatalogService(_service.service): ...

class UnityCatalogService_Stub(UnityCatalogService): ...
